#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019

"""
Configurations.

configuration file looking for path:
    1. $IDDS_CONFIG
    2. $IDDS_HOME/etc/idds/idds.cfg
    3. /etc/idds/idds.cfg
    4. $VIRTUAL_ENV/etc/idds/idds.cfg
"""


import os

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser


def is_client():
    if 'IDDS_CLIENT_MODE' in os.environ and os.environ['IDDS_CLIENT_MODE']:
        client_mode = True
    else:
        client_mode = False
        try:
            import idds.agents.common                # noqa F401
        except ModuleNotFoundError as ex:            # noqa F841
            client_mode = True

    return client_mode


def config_has_section(section):
    """
    Return where there is a section

    :param section: the named section.
.
    :returns: True/False.
    """
    __CONFIG = get_config()
    return __CONFIG.has_section(section)


def config_has_option(section, option):
    """
    Return where there is an option for a given option in a section

    :param section: the named section.
    :param option: the named option.
.
    :returns: True/False.
    """
    __CONFIG = get_config()
    return __CONFIG.has_option(section, option)


def config_list_options(section):
    """
    Return list of (name, value) for a given option in a section

    :param section: the named section.
.
    :returns: list of (name, value).
    """
    __CONFIG = get_config()
    return __CONFIG.items(section)


def config_get(section, option):
    """
    Return the string value for a given option in a section
    :param section: the named section.
    :param option: the named option.
.
    :returns: the configuration value.
    """
    __CONFIG = get_config()
    return __CONFIG.get(section, option)


def config_get_int(section, option):
    """
    Return the integer value for a given option in a section
    :param section: the named section.
    :param option: the named option.
.
    :returns: the integer configuration value.
    """
    __CONFIG = get_config()
    return __CONFIG.getint(section, option)


def config_get_float(section, option):
    """
    Return the float value for a given option in a section
    :param section: the named section.
    :param option: the named option.
.
    :returns: the float configuration value.
    """
    __CONFIG = get_config()
    return __CONFIG.getfloat(section, option)


def config_get_bool(section, option):
    """
    Return the boolean value for a given option in a section
    :param section: the named section.
    :param option: the named option.
.
    :returns: the boolean configuration value.
    """
    __CONFIG = get_config()
    return __CONFIG.getboolean(section, option)


def get_local_config_root(local_config_root=None):
    if 'IDDS_LOCAL_CONFIG_ROOT' in os.environ and os.environ['IDDS_LOCAL_CONFIG_ROOT']:
        if local_config_root is None:
            # print("IDDS_LOCAL_CONFIG_ROOT is set. Will use it.")
            local_config_root = os.environ['IDDS_LOCAL_CONFIG_ROOT']
        else:
            print("local_config_root is set to %s. Ignore IDDS_LOCAL_CONFIG_ROOT" % local_config_root)

    if local_config_root is None:
        # local_config_root = "~/.idds"
        home_dir = os.path.expanduser("~")
        if os.access(home_dir, os.W_OK):
            local_config_root = os.path.join(home_dir, ".idds")
        else:
            # username = os.getlogin()
            # if username == 'root':
            #     local_config_root = os.path.join('/tmp', ".idds")
            # else:
            #     local_config_root = os.path.join(os.path.join("/tmp", username), ".idds")
            pass

    if local_config_root and not os.path.exists(local_config_root):
        try:
            os.makedirs(local_config_root, exist_ok=True)
        except Exception as ex:
            print("Failed to create %s: %s", local_config_root, str(ex))
            local_config_root = None
    if local_config_root and not os.access(local_config_root, os.W_OK):
        print("No write permission on local config root %s, set it to None" % (local_config_root))
        local_config_root = None
    return local_config_root


def get_local_cfg_file(local_config_root=None):
    local_config_root = get_local_config_root(local_config_root)
    if local_config_root:
        local_cfg = os.path.join(local_config_root, 'idds_local.cfg')
        return local_cfg
    return None


def get_local_config_value(configuration, section, name, current, default):
    value = None
    if configuration and configuration.has_section(section) and configuration.has_option(section, name):
        if name in ['oidc_refresh_lifetime']:
            value = configuration.getint(section, name)
        elif name in ['oidc_auto', 'oidc_polling']:
            value = configuration.getboolean(section, name)
        else:
            value = configuration.get(section, name)
    if current is not None:
        value = current
    elif value is None:
        value = default
    return value


def get_main_config_file():
    __CONFIG = ConfigParser.ConfigParser()
    if os.environ.get('IDDS_CONFIG', None):
        configfile = os.environ['IDDS_CONFIG']
        if __CONFIG.read(configfile) == [configfile]:
            return configfile
    else:
        configfiles = ['%s/etc/idds/idds.cfg' % os.environ.get('IDDS_HOME', ''),
                       '/etc/idds/idds.cfg',
                       '%s/etc/idds/idds.cfg' % os.environ.get('VIRTUAL_ENV', '')]
        for configfile in configfiles:
            if __CONFIG.read(configfile) == [configfile]:
                return configfile
    return None


def get_main_config():
    __CONFIG = ConfigParser.ConfigParser()

    __HAS_CONFIG = False
    if os.environ.get('IDDS_CONFIG', None):
        configfile = os.environ['IDDS_CONFIG']
        if not __CONFIG.read(configfile) == [configfile]:
            raise Exception('IDDS_CONFIG is defined as %s, ' % configfile,
                            'but could not load configurations from it.')
        __HAS_CONFIG = True
    else:
        configfiles = ['%s/etc/idds/idds.cfg' % os.environ.get('IDDS_HOME', ''),
                       '/etc/idds/idds.cfg',
                       '%s/etc/idds/idds.cfg' % os.environ.get('VIRTUAL_ENV', '')]

        for configfile in configfiles:
            if __CONFIG.read(configfile) == [configfile]:
                __HAS_CONFIG = True
                # print("Configuration file %s is used" % configfile)
                break
    return __CONFIG, __HAS_CONFIG


def get_config():
    __CONFIG, __HAS_CONFIG = get_main_config()

    if not __HAS_CONFIG:
        local_cfg = get_local_cfg_file()
        if os.path.exists(local_cfg):
            __CONFIG.read(local_cfg)
            __HAS_CONFIG = True
        else:
            if not is_client():
                raise Exception("Could not load configuration file."
                                "For iDDS client, please run 'idds setup' to create local config file."
                                "For an iDDS server, IDDS looks for a configuration file, in order:"
                                "\n\t${IDDS_CONFIG}"
                                "\n\t${IDDS_HOME}/etc/idds/idds.cfg"
                                "\n\t/etc/idds/idds.cfg"
                                "\n\t${VIRTUAL_ENV}/etc/idds/idds.cfg")
    return __CONFIG



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
plugin utils

plugin configuration structure:

[section_name]
plugin_sequence = <plugin_name1>,<plugin_name2>
plugin.<plugin_name> = <plugin.lib.path.ClassName>
plugin.<plugin_name>.<attr1> = <value1>
plugin.<plugin_name>.<attr2> = <value2>
"""


from idds.common.config import (config_has_section, config_has_option,
                                config_list_options, config_get)


def load_plugin_sequence(config_section, config_option='plugin_sequence'):
    """
    load plugin sequence
    """
    plugin_sequence = []
    if config_has_section(config_section) and config_has_option(config_section, config_option):
        plugin_sequence = config_get(config_section, config_option)
        plugin_sequence = plugin_sequence.split(",")
        plugin_sequence = [plugin.strip() for plugin in plugin_sequence]
    return plugin_sequence


def load_plugin_attributes(config_section, name, plugin):
    """
    Load plugin attributes
    """
    attrs = {}
    if config_has_section(config_section):
        options = config_list_options(config_section)
        for option, value in options:
            plugin_prefix = 'plugin.%s.' % name
            if option.startswith(plugin_prefix):
                attr_name = option.replace(plugin_prefix, '')
                if isinstance(value, str) and value.lower() == 'true':
                    value = True
                if isinstance(value, str) and value.lower() == 'false':
                    value = False
                attrs[attr_name] = value
    return attrs


def load_plugin(config_section, name, plugin):
    """
    Load plugin attributes
    """
    attrs = load_plugin_attributes(config_section, name, plugin)
    k = plugin.rfind('.')
    plugin_modules = plugin[:k]
    plugin_class = plugin[k + 1:]
    module = __import__(plugin_modules, fromlist=[None])
    cls = getattr(module, plugin_class)
    impl = cls(**attrs)
    return impl


def load_plugins(config_section):
    """
    Load plugins
    """
    plugins = {}
    if config_has_section(config_section):
        options = config_list_options(config_section)
        for option, value in options:
            if option.startswith('plugin.'):
                if option.count('.') == 1:
                    plugin_name = option.replace('plugin.', '').strip()
                    plugins[plugin_name] = load_plugin(config_section, plugin_name, value)
    return plugins



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2022

"""
Constants.
"""

from enum import Enum


SCOPE_LENGTH = 25
NAME_LENGTH = 255
LONG_NAME_LENGTH = 4000


class Sections:
    Main = 'main'
    Common = 'common'
    Clerk = 'clerk'
    Marshaller = 'marshaller'
    Transformer = 'transformer'
    Transporter = 'transporter'
    Carrier = 'carrier'
    Conductor = 'conductor'
    Consumer = 'consumer'
    EventBus = 'eventbus'
    Cache = 'cache'


class HTTP_STATUS_CODE:
    OK = 200
    Created = 201
    Accepted = 202

    # Client Errors
    BadRequest = 400
    Unauthorized = 401
    Forbidden = 403
    NotFound = 404
    NoMethod = 405
    Conflict = 409

    # Server Errors
    InternalError = 500


class IDDSEnum(Enum):
    def to_dict(self):
        ret = {'class': self.__class__.__name__,
               'module': self.__class__.__module__,
               'attributes': {}}
        for key, value in self.__dict__.items():
            if not key.startswith('__'):
                if key == 'logger':
                    value = None
                if value and hasattr(value, 'to_dict'):
                    value = value.to_dict()
                ret['attributes'][key] = value
        return ret

    @staticmethod
    def is_class(d):
        if d and isinstance(d, dict) and 'class' in d and 'module' in d and 'attributes' in d:
            return True
        return False

    @staticmethod
    def load_instance(d):
        module = __import__(d['module'], fromlist=[None])
        cls = getattr(module, d['class'])
        if issubclass(cls, Enum):
            impl = cls(d['attributes']['_value_'])
        else:
            impl = cls()
        return impl

    @staticmethod
    def from_dict(d):
        if IDDSEnum.is_class(d):
            impl = IDDSEnum.load_instance(d)
            for key, value in d['attributes'].items():
                if key == 'logger':
                    continue
                if IDDSEnum.is_class(value):
                    value = IDDSEnum.from_dict(value)
                setattr(impl, key, value)
            return impl
        return d


class WorkStatus(IDDSEnum):
    New = 0
    Ready = 1
    Transforming = 2
    Finished = 3
    SubFinished = 4
    Failed = 5
    Extend = 6
    ToCancel = 7
    Cancelling = 8
    Cancelled = 9
    ToSuspend = 10
    Suspending = 11
    Suspended = 12
    ToResume = 13
    Resuming = 14
    ToExpire = 15
    Expiring = 16
    Expired = 17
    ToFinish = 18
    ToForceFinish = 19
    Running = 20
    Terminating = 21


class RequestStatus(IDDSEnum):
    New = 0
    Ready = 1
    Transforming = 2
    Finished = 3
    SubFinished = 4
    Failed = 5
    Extend = 6
    ToCancel = 7
    Cancelling = 8
    Cancelled = 9
    ToSuspend = 10
    Suspending = 11
    Suspended = 12
    ToResume = 13
    Resuming = 14
    ToExpire = 15
    Expiring = 16
    Expired = 17
    ToFinish = 18
    ToForceFinish = 19
    Terminating = 20


class RequestLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class WorkprogressStatus(IDDSEnum):
    New = 0
    Ready = 1
    Transforming = 2
    Finished = 3
    SubFinished = 4
    Failed = 5
    Extend = 6
    ToCancel = 7
    Cancelling = 8
    Cancelled = 9
    ToSuspend = 10
    Suspending = 11
    Suspended = 12
    ToResume = 13
    Resuming = 14
    ToExpire = 15
    Expiring = 16
    Expired = 17
    ToFinish = 18
    ToForceFinish = 19


class WorkprogressLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class RequestType(IDDSEnum):
    Workflow = 0
    EventStreaming = 1
    StageIn = 2
    ActiveLearning = 3
    HyperParameterOpt = 4
    Derivation = 5
    Other = 99


class TransformType(IDDSEnum):
    Workflow = 0
    EventStreaming = 1
    StageIn = 2
    ActiveLearning = 3
    HyperParameterOpt = 4
    Derivation = 5
    Processing = 6
    Actuating = 7
    Data = 8
    Other = 99


class TransformStatus(IDDSEnum):
    New = 0
    Ready = 1
    Transforming = 2
    Finished = 3
    SubFinished = 4
    Failed = 5
    Extend = 6
    ToCancel = 7
    Cancelling = 8
    Cancelled = 9
    ToSuspend = 10
    Suspending = 11
    Suspended = 12
    ToResume = 13
    Resuming = 14
    ToExpire = 15
    Expiring = 16
    Expired = 17
    ToFinish = 18
    ToForceFinish = 19
    Terminating = 20


class TransformLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class CollectionType(IDDSEnum):
    Container = 0
    Dataset = 1
    File = 2
    PseudoDataset = 3


class CollectionRelationType(IDDSEnum):
    Input = 0
    Output = 1
    Log = 2


class CollectionStatus(IDDSEnum):
    New = 0
    Updated = 1
    Processing = 2
    Open = 3
    Closed = 4
    SubClosed = 5
    Failed = 6
    Deleted = 7
    Cancelled = 8
    Suspended = 9


class CollectionLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class ContentType(IDDSEnum):
    File = 0
    Event = 1
    PseudoContent = 2


class ContentRelationType(IDDSEnum):
    Input = 0
    Output = 1
    Log = 2
    InputDependency = 3


class ContentStatus(IDDSEnum):
    New = 0
    Processing = 1
    Available = 2
    Failed = 3
    FinalFailed = 4
    Lost = 5
    Deleted = 6
    Mapped = 7
    FakeAvailable = 8
    Missing = 9


class ContentLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class GranularityType(IDDSEnum):
    File = 0
    Event = 1


class ProcessingStatus(IDDSEnum):
    New = 0
    Submitting = 1
    Submitted = 2
    Running = 3
    Finished = 4
    Failed = 5
    Lost = 6
    Cancel = 7
    FinishedOnStep = 8
    FinishedOnExec = 9
    FinishedTerm = 10
    SubFinished = 11
    ToCancel = 12
    Cancelling = 13
    Cancelled = 14
    ToSuspend = 15
    Suspending = 16
    Suspended = 17
    ToResume = 18
    Resuming = 19
    ToExpire = 20
    Expiring = 21
    Expired = 22
    TimeOut = 23
    ToFinish = 24
    ToForceFinish = 25
    Broken = 26
    Terminating = 27
    ToTrigger = 28
    Triggering = 29


class ProcessingLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class MessageType(IDDSEnum):
    StageInFile = 0
    StageInCollection = 1
    StageInWork = 2
    ActiveLearningFile = 3
    ActiveLearningCollection = 4
    ActiveLearningWork = 5
    HyperParameterOptFile = 6
    HyperParameterOptCollection = 7
    HyperParameterOptWork = 8
    ProcessingFile = 9
    ProcessingCollection = 10
    ProcessingWork = 11
    HealthHeartbeat = 12
    IDDSCommunication = 13
    UnknownFile = 97
    UnknownCollection = 98
    UnknownWork = 99


class MessageTypeStr(IDDSEnum):
    StageInFile = 'file_stagein'
    StageInCollection = 'collection_stagein'
    StageInWork = 'work_stagein'
    ActiveLearningFile = 'file_activelearning'
    ActiveLearningCollection = 'collection_activelearning'
    ActiveLearningWork = 'work_activelearning'
    HyperParameterOptFile = 'file_hyperparameteropt'
    HyperParameterOptCollection = 'collection_hyperparameteropt'
    HyperParameterOptWork = 'work_hyperparameteropt'
    ProcessingFile = 'file_processing'
    ProcessingCollection = 'collection_processing'
    ProcessingWork = 'work_processing'
    HealthHeartbeat = 'health_heartbeat'
    IDDSCommunication = 'idds_communication'
    UnknownFile = 'file_unknown'
    UnknownCollection = 'collection_unknown'
    UnknownWork = 'work_unknown'


TransformType2MessageTypeMap = {
    '0': {'transform_type': TransformType.Workflow,
          'work': (MessageType.UnknownWork, MessageTypeStr.UnknownWork),
          'collection': (MessageType.UnknownCollection, MessageTypeStr.UnknownCollection),
          'file': (MessageType.UnknownFile, MessageTypeStr.UnknownFile)
          },
    '2': {'transform_type': TransformType.StageIn,
          'work': (MessageType.StageInWork, MessageTypeStr.StageInWork),
          'collection': (MessageType.StageInCollection, MessageTypeStr.StageInCollection),
          'file': (MessageType.StageInFile, MessageTypeStr.StageInFile)
          },
    '3': {'transform_type': TransformType.ActiveLearning,
          'work': (MessageType.ActiveLearningWork, MessageTypeStr.ActiveLearningWork),
          'collection': (MessageType.ActiveLearningCollection, MessageTypeStr.ActiveLearningCollection),
          'file': (MessageType.ActiveLearningFile, MessageTypeStr.ActiveLearningFile)
          },
    '4': {'transform_type': TransformType.HyperParameterOpt,
          'work': (MessageType.HyperParameterOptWork, MessageTypeStr.HyperParameterOptWork),
          'collection': (MessageType.HyperParameterOptCollection, MessageTypeStr.HyperParameterOptCollection),
          'file': (MessageType.HyperParameterOptFile, MessageTypeStr.HyperParameterOptFile)
          },
    '6': {'transform_type': TransformType.Processing,
          'work': (MessageType.ProcessingWork, MessageTypeStr.ProcessingWork),
          'collection': (MessageType.ProcessingCollection, MessageTypeStr.ProcessingCollection),
          'file': (MessageType.ProcessingFile, MessageTypeStr.ProcessingFile)
          }
}


class MessageStatus(IDDSEnum):
    New = 0
    Fetched = 1
    Delivered = 2
    Failed = 3


class MessageLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class MessageSource(IDDSEnum):
    Clerk = 0
    Transformer = 1
    Transporter = 2
    Carrier = 3
    Conductor = 4
    Rest = 5


class MessageDestination(IDDSEnum):
    Clerk = 0
    Transformer = 1
    Transporter = 2
    Carrier = 3
    Conductor = 4
    Outside = 5


class CommandType(IDDSEnum):
    AbortRequest = 0
    ResumeRequest = 1
    ExpireRequest = 2


class CommandStatus(IDDSEnum):
    New = 0
    Processing = 1
    Processed = 2
    Failed = 3
    UnknownCommand = 4


class CommandLocking(IDDSEnum):
    Idle = 0
    Locking = 1


class CommandLocation(IDDSEnum):
    Clerk = 0
    Transformer = 1
    Transporter = 2
    Carrier = 3
    Conductor = 4
    Rest = 5
    Other = 6


def get_work_status_from_transform_processing_status(status):
    if status in [ProcessingStatus.New, TransformStatus.New]:
        return WorkStatus.New
    elif status in [ProcessingStatus.Submitting, ProcessingStatus.Submitted, TransformStatus.Transforming]:
        return WorkStatus.Transforming
    elif status in [ProcessingStatus.Running]:
        return WorkStatus.Transforming
    elif status in [ProcessingStatus.Finished, TransformStatus.Finished]:
        return WorkStatus.Finished
    elif status in [ProcessingStatus.Failed, ProcessingStatus.Broken, TransformStatus.Failed]:
        return WorkStatus.Failed
    elif status in [ProcessingStatus.SubFinished, TransformStatus.SubFinished]:
        return WorkStatus.SubFinished
    elif status in [ProcessingStatus.Cancelled, ProcessingStatus.Suspended, TransformStatus.Cancelled, TransformStatus.Suspended]:
        return WorkStatus.Cancelled
    elif status in [ProcessingStatus.Terminating, TransformStatus.Terminating]:
        return WorkStatus.Terminating
    else:
        return WorkStatus.Transforming



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2020


from idds.common.constants import (WorkprogressStatus)
from idds.common.utils import is_sub


def get_workprogresses_status(wp_status):
    if not wp_status:
        return False

    if not isinstance(wp_status, (list, tuple)):
        wp_status = [wp_status]
    if len(wp_status) == 1:
        return wp_status[0]
    elif is_sub(wp_status, [WorkprogressStatus.Finished]):
        return WorkprogressStatus.Finished
    elif is_sub(wp_status, [WorkprogressStatus.Finished, WorkprogressStatus.SubFinished]):
        return WorkprogressStatus.SubFinished
    elif is_sub(wp_status, [WorkprogressStatus.Finished, WorkprogressStatus.SubFinished,
                            WorkprogressStatus.Failed]):
        return WorkprogressStatus.Failed
    elif is_sub(wp_status, [WorkprogressStatus.Finished, WorkprogressStatus.SubFinished,
                            WorkprogressStatus.Failed, WorkprogressStatus.Cancelled]):
        return WorkprogressStatus.Cancelled
    elif is_sub(wp_status, [WorkprogressStatus.Finished, WorkprogressStatus.SubFinished,
                            WorkprogressStatus.Failed, WorkprogressStatus.Cancelled,
                            WorkprogressStatus.Transforming]):
        return WorkprogressStatus.Transforming
    elif is_sub(wp_status, [WorkprogressStatus.Finished, WorkprogressStatus.SubFinished,
                            WorkprogressStatus.Failed, WorkprogressStatus.Cancelled,
                            WorkprogressStatus.Transforming, WorkprogressStatus.Cancelling]):
        return WorkprogressStatus.Cancelling
    else:
        return WorkprogressStatus.Transforming



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

"""
Dict class.
"""

import inspect
from enum import Enum


class DictClass(object):
    def to_dict_l(self, d):
        # print('to_dict_l')
        # print(d)
        if not d:
            return d

        if hasattr(d, 'to_dict'):
            return d.to_dict()
        elif isinstance(d, dict):
            new_d = {}
            for k, v in d.items():
                new_d[k] = self.to_dict_l(v)
            return new_d
        elif isinstance(d, list):
            new_d = []
            for k in d:
                new_d.append(self.to_dict_l(k))
            return new_d
        elif inspect.ismethod(d):
            return {'idds_method': d.__name__, 'idds_method_class_id': d.__self__.get_internal_id()}
        return d

    def to_dict(self):
        # print('to_dict')
        ret = {'class': self.__class__.__name__,
               'module': self.__class__.__module__,
               'attributes': {}}
        for key, value in self.__dict__.items():
            # print(key)
            # print(value)
            # if not key.startswith('__') and not key.startswith('_'):
            if not key.startswith('__'):
                if key == 'logger':
                    new_value = None
                else:
                    new_value = self.to_dict_l(value)
                ret['attributes'][key] = new_value
        return ret

    @staticmethod
    def is_class(d):
        if d and isinstance(d, dict) and 'class' in d and 'module' in d and 'attributes' in d:
            return True
        return False

    @staticmethod
    def is_class_method(d):
        if d and isinstance(d, dict) and 'idds_method' in d and 'idds_method_class_id' in d:
            return True
        return False

    @staticmethod
    def is_class_attribute(d):
        if d and isinstance(d, dict) and 'idds_attribute' in d and 'idds_method_class_id' in d:
            return True
        return False

    @staticmethod
    def load_instance(d):
        module = __import__(d['module'], fromlist=[None])
        cls = getattr(module, d['class'])
        if issubclass(cls, Enum):
            impl = cls(d['attributes']['_value_'])
        else:
            impl = cls()
        return impl

    @staticmethod
    def load_instance_method(d):
        # not do anything. Will load the method in Workflow class.
        return d

    @staticmethod
    def load_instance_attribute(d):
        # not do anything. Will load the method in Workflow class.
        return d

    @staticmethod
    def from_dict(d):
        if not d:
            return d

        if DictClass.is_class(d):
            impl = DictClass.load_instance(d)
            for key, value in d['attributes'].items():
                if key == 'logger':
                    continue
                # elif key == 'output_data':
                #     continue
                else:
                    value = DictClass.from_dict(value)
                setattr(impl, key, value)
            return impl
        elif DictClass.is_class_method(d):
            impl = DictClass.load_instance_method(d)
            return impl
        elif DictClass.is_class_attribute(d):
            impl = DictClass.load_instance_attribute(d)
            return impl
        elif isinstance(d, dict):
            for k, v in d.items():
                d[k] = DictClass.from_dict(v)
            return d
        elif isinstance(d, list):
            new_d = []
            for k in d:
                new_d.append(DictClass.from_dict(k))
            return new_d
        else:
            return d

        return d



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@@cern.ch>, 2021 - 2022

import datetime
import base64
import json
import jwt
import os
import re
import requests

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

try:
    # Python 2
    from urllib import urlencode
except ImportError:
    # Python 3
    from urllib.parse import urlencode
    raw_input = input

# from cryptography import x509
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

# from idds.common import exceptions
from idds.common.constants import HTTP_STATUS_CODE


def decode_value(val):
    if isinstance(val, str):
        val = val.encode()
    decoded = base64.urlsafe_b64decode(val + b'==')
    return int.from_bytes(decoded, 'big')


def should_verify():
    if os.environ.get('IDDS_AUTH_NO_VERIFY', None):
        return False
    return True


class BaseAuthentication(object):
    def __init__(self, timeout=None):
        self.timeout = timeout
        self.config = self.load_auth_server_config()
        self.max_expires_in = 60
        if self.config and self.config.has_section('common'):
            if self.config.has_option('common', 'max_expires_in'):
                self.max_expires_in = self.config.getint('common', 'max_expires_in')

    def load_auth_server_config(self):
        config = ConfigParser.ConfigParser()
        if os.environ.get('IDDS_AUTH_CONFIG', None):
            configfile = os.environ['IDDS_AUTH_CONFIG']
            if config.read(configfile) == [configfile]:
                return config

        configfiles = ['%s/etc/idds/auth/auth.cfg' % os.environ.get('IDDS_HOME', ''),
                       '/etc/idds/auth/auth.cfg', '/opt/idds/etc/idds/auth/auth.cfg',
                       '%s/etc/idds/auth/auth.cfg' % os.environ.get('VIRTUAL_ENV', '')]
        for configfile in configfiles:
            if config.read(configfile) == [configfile]:
                return config
        return config

    def get_allow_vos(self):
        section = 'common'
        allow_vos = []
        if self.config and self.config.has_section(section):
            if self.config.has_option(section, 'allow_vos'):
                allow_vos_temp = self.config.get(section, 'allow_vos')
                allow_vos_temp = allow_vos_temp.split(',')
                for t in allow_vos_temp:
                    t = t.strip()
                    allow_vos.append(t)
        return allow_vos


class OIDCAuthentication(BaseAuthentication):
    def __init__(self, timeout=None):
        super(OIDCAuthentication, self).__init__(timeout=timeout)

    def get_auth_config(self, vo):
        ret = {'vo': vo, 'oidc_config_url': None, 'client_id': None,
               'client_secret': None, 'audience': None}

        if self.config and self.config.has_section(vo):
            for name in ['oidc_config_url', 'client_id', 'client_secret', 'vo', 'audience']:
                if self.config.has_option(vo, name):
                    ret[name] = self.config.get(vo, name)
        return ret

    def get_http_content(self, url):
        try:
            r = requests.get(url, allow_redirects=True, verify=should_verify())
            return r.content
        except Exception as error:
            return False, 'Failed to get http content for %s: %s' (str(url), str(error))

    def get_endpoint_config(self, auth_config):
        content = self.get_http_content(auth_config['oidc_config_url'])
        endpoint_config = json.loads(content)
        # ret = {'token_endpoint': , 'device_authorization_endpoint': None}
        return endpoint_config

    def get_oidc_sign_url(self, vo):
        try:
            allow_vos = self.get_allow_vos()
            if vo not in allow_vos:
                return False, "VO %s is not allowed." % vo

            auth_config = self.get_auth_config(vo)
            endpoint_config = self.get_endpoint_config(auth_config)

            data = {'client_id': auth_config['client_id'],
                    'scope': "openid profile email offline_access",
                    'audience': auth_config['audience']}

            headers = {'content-type': 'application/x-www-form-urlencoded'}

            result = requests.session().post(endpoint_config['device_authorization_endpoint'],
                                             # data=json.dumps(data),
                                             urlencode(data).encode(),
                                             timeout=self.timeout,
                                             verify=should_verify(),
                                             headers=headers)

            if result is not None:
                if result.status_code == HTTP_STATUS_CODE.OK and result.text:
                    return True, json.loads(result.text)
                else:
                    return False, "Failed to get oidc sign in URL (status: %s, text: %s)" % (result.status_code, result.text)
            else:
                return False, "Failed to get oidc sign in URL. Response is None."
        except requests.exceptions.ConnectionError as error:
            return False, 'Failed to get oidc sign in URL. ConnectionError: ' + str(error)
        except Exception as error:
            return False, 'Failed to get oidc sign in URL: ' + str(error)

    def get_id_token(self, vo, device_code, interval=5, expires_in=60):
        try:
            allow_vos = self.get_allow_vos()
            if vo not in allow_vos:
                return False, "VO %s is not allowed." % vo

            auth_config = self.get_auth_config(vo)
            endpoint_config = self.get_endpoint_config(auth_config)

            data = {'client_id': auth_config['client_id'],
                    'client_secret': auth_config['client_secret'],
                    'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
                    'device_code': device_code}

            headers = {'content-type': 'application/x-www-form-urlencoded'}

            if not interval:
                interval = 5
            interval = int(interval)

            if not expires_in:
                expires_in = 60
            expires_in = int(expires_in)
            if expires_in > self.max_expires_in:
                expires_in = self.max_expires_in

            result = requests.session().post(endpoint_config['token_endpoint'],
                                             # data=json.dumps(data),
                                             urlencode(data).encode(),
                                             timeout=self.timeout,
                                             verify=should_verify(),
                                             headers=headers)
            if result is not None:
                if result.status_code == HTTP_STATUS_CODE.OK and result.text:
                    return True, json.loads(result.text)
                else:
                    return False, json.loads(result.text)
            else:
                return False, None
        except Exception as error:
            return False, 'Failed to get oidc token: ' + str(error)

    def refresh_id_token(self, vo, refresh_token):
        try:
            allow_vos = self.get_allow_vos()
            if vo not in allow_vos:
                return False, "VO %s is not allowed." % vo

            auth_config = self.get_auth_config(vo)
            endpoint_config = self.get_endpoint_config(auth_config)

            data = {'client_id': auth_config['client_id'],
                    'client_secret': auth_config['client_secret'],
                    'grant_type': 'refresh_token',
                    'refresh_token': refresh_token}

            headers = {'content-type': 'application/x-www-form-urlencoded'}

            result = requests.session().post(endpoint_config['token_endpoint'],
                                             # data=json.dumps(data),
                                             urlencode(data).encode(),
                                             timeout=self.timeout,
                                             verify=should_verify(),
                                             headers=headers)

            if result is not None:
                if result.status_code == HTTP_STATUS_CODE.OK and result.text:
                    return True, json.loads(result.text)
                else:
                    return False, "Failed to refresh oidc token (status: %s, text: %s)" % (result.status_code, result.text)
            else:
                return False, "Failed to refresh oidc token. Response is None."
        except requests.exceptions.ConnectionError as error:
            return False, 'Failed to refresh oidc token. ConnectionError: ' + str(error)
        except Exception as error:
            return False, 'Failed to refresh oidc token: ' + str(error)

    def get_public_key(self, token, jwks_uri):
        headers = jwt.get_unverified_header(token)
        if headers is None or 'kid' not in headers:
            raise jwt.exceptions.InvalidTokenError('cannot extract kid from headers')
        kid = headers['kid']

        jwks_content = self.get_http_content(jwks_uri)
        jwks = json.loads(jwks_content)
        jwk = None
        for j in jwks.get('keys', []):
            if j.get('kid') == kid:
                jwk = j
        if jwk is None:
            raise jwt.exceptions.InvalidTokenError('JWK not found for kid={0}: {1}'.format(kid, str(jwks)))

        public_num = RSAPublicNumbers(n=decode_value(jwk['n']), e=decode_value(jwk['e']))
        public_key = public_num.public_key(default_backend())
        pem = public_key.public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)
        return pem

    def verify_id_token(self, vo, token):
        try:
            allow_vos = self.get_allow_vos()
            if vo not in allow_vos:
                return False, "VO %s is not allowed." % vo, None

            auth_config = self.get_auth_config(vo)
            endpoint_config = self.get_endpoint_config(auth_config)

            # check audience
            decoded_token = jwt.decode(token, verify=False, options={"verify_signature": False})
            audience = decoded_token['aud']
            if auth_config['client_id'] != audience:
                # discovery_endpoint = auth_config['oidc_config_url']
                return False, "The audience of the token doesn't match vo configuration.", None

            public_key = self.get_public_key(token, endpoint_config['jwks_uri'])
            # decode token only with RS256
            decoded = jwt.decode(token, public_key, verify=True, algorithms='RS256',
                                 audience=audience, issuer=endpoint_config['issuer'])
            decoded['vo'] = vo
            if 'name' in decoded:
                username = decoded['name']
            else:
                username = None
            return True, decoded, username
        except Exception as error:
            return False, 'Failed to verify oidc token: ' + str(error), None


class OIDCAuthenticationUtils(object):
    def __init__(self):
        pass

    def save_token(self, path, token):
        try:
            with open(path, 'w') as f:
                f.write(json.dumps(token))
            return True, None
        except Exception as error:
            return False, "Failed to save token: %s" % str(error)

    def load_token(self, path):
        try:
            with open(path) as f:
                data = json.load(f)
            return True, data
        except Exception as error:
            return False, "Failed to load token: %s" % str(error)

    def is_token_expired(self, token):
        try:
            enc = token['id_token'].split('.')[1]
            enc += '=' * (-len(enc) % 4)
            dec = json.loads(base64.urlsafe_b64decode(enc.encode()))
            exp_time = datetime.datetime.utcfromtimestamp(dec['exp'])
            # delta = exp_time - datetime.datetime.utcnow()
            if exp_time < datetime.datetime.utcnow():
                return True, None
            else:
                return False, None
        except Exception as error:
            return True, "Failed to parse token: %s" % str(error)

    def clean_token(self, path):
        try:
            os.remove(path)
            return True, None
        except Exception as error:
            return False, "Failed to clean token: %s" % str(error)

    def get_token_info(self, token):
        try:
            enc = token['id_token'].split('.')[1]
            enc += '=' * (-len(enc) % 4)
            dec = json.loads(base64.urlsafe_b64decode(enc.encode()))
            exp_time = datetime.datetime.utcfromtimestamp(dec['exp'])
            delta = exp_time - datetime.datetime.utcnow()
            minutes = delta.total_seconds() / 60

            info = dec
            info['expire'] = exp_time
            info['expire_time'] = 'Token will expire in %s minutes' % minutes
            info['expire_at'] = 'Token will expire at {0} UTC'.format(exp_time.strftime("%Y-%m-%d %H:%M:%S"))
            return True, info
        except Exception as error:
            return True, "Failed to parse token: %s" % str(error)


class X509Authentication(BaseAuthentication):
    def __init__(self, timeout=None):
        super(X509Authentication, self).__init__(timeout=timeout)

    def get_ban_user_list(self):
        section = "Users"
        option = "ban_users"
        if self.config and self.config.has_section(section):
            if self.config.has_option(section, option):
                users = self.config.get(section, option)
                users = users.split(",")
                return users
        return []

    def get_allow_user_list(self):
        section = "Users"
        option = "allow_users"
        if self.config and self.config.has_section(section):
            if self.config.has_option(section, option):
                users = self.config.get(section, option)
                users = users.split(",")
                return users
        return []

    def get_super_user_list(self):
        section = "Users"
        option = "super_users"
        if self.config and self.config.has_section(section):
            if self.config.has_option(section, option):
                users = self.config.get(section, option)
                users = users.split(",")
                return users
        return []


# "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=wguan/CN=667815/CN=Wen Guan/CN=1883443395"
def get_user_name_from_dn1(dn):
    try:
        up = re.compile('/(DC|O|OU|C|L)=[^\/]+')        # noqa W605
        username = up.sub('', dn)
        up2 = re.compile('/CN=[0-9]+')
        username = up2.sub('', username)
        up3 = re.compile(' [0-9]+')
        username = up3.sub('', username)
        up4 = re.compile('_[0-9]+')
        username = up4.sub('', username)
        username = username.replace('/CN=proxy', '')
        username = username.replace('/CN=limited proxy', '')
        username = username.replace('limited proxy', '')
        username = re.sub('/CN=Robot:[^/]+', '', username)
        username = re.sub('/CN=nickname:[^/]+', '', username)
        pat = re.compile('.*/CN=([^\/]+)/CN=([^\/]+)')         # noqa W605
        mat = pat.match(username)
        if mat:
            username = mat.group(2)
        else:
            username = username.replace('/CN=', '')
        if username.lower().find('/email') > 0:
            username = username[:username.lower().find('/email')]
        pat = re.compile('.*(limited.*proxy).*')
        mat = pat.match(username)
        if mat:
            username = mat.group(1)
        username = username.replace('(', '')
        username = username.replace(')', '')
        username = username.replace("'", '')
        return username
    except Exception:
        return dn


# 'CN=203633261,CN=Wen Guan,CN=667815,CN=wguan,OU=Users,OU=Organic Units,DC=cern,DC=ch'
def get_user_name_from_dn2(dn):
    try:
        up = re.compile(',(DC|O|OU|C|L)=[^\,]+')        # noqa W605
        username = up.sub('', dn)
        up2 = re.compile(',CN=[0-9]+')
        username = up2.sub('', username)
        up3 = re.compile(' [0-9]+')
        username = up3.sub('', username)
        up4 = re.compile('_[0-9]+')
        username = up4.sub('', username)
        username = username.replace(',CN=proxy', '')
        username = username.replace(',CN=limited proxy', '')
        username = username.replace('limited proxy', '')
        username = re.sub(',CN=Robot:[^/]+', '', username)
        username = re.sub(',CN=nickname:[^/]+', '', username)
        pat = re.compile('.*,CN=([^\,]+),CN=([^\,]+)')         # noqa W605
        mat = pat.match(username)
        if mat:
            username = mat.group(1)
        else:
            username = username.replace(',CN=', '')
        if username.lower().find(',email') > 0:
            username = username[:username.lower().find(',email')]
        pat = re.compile('.*(limited.*proxy).*')
        mat = pat.match(username)
        if mat:
            username = mat.group(1)
        username = username.replace('(', '')
        username = username.replace(')', '')
        username = username.replace("'", '')
        return username
    except Exception:
        return dn


def get_user_name_from_dn(dn):
    dn = get_user_name_from_dn1(dn)
    dn = get_user_name_from_dn2(dn)
    return dn


def authenticate_x509(vo, dn, client_cert):
    if not dn:
        return False, "User DN cannot be found.", None
    if not client_cert:
        return False, "Client certificate proxy cannot be found.", None

    # certDecoded = x509.load_pem_x509_certificate(str.encode(client_cert), default_backend())
    # print(certDecoded.issuer)
    # for ext in certDecoded.extensions:
    #     print(ext)
    allow_user_list = X509Authentication().get_allow_user_list()
    matched = False
    for allow_user in allow_user_list:
        # pat = re.compile(allow_user)
        # mat = pat.match(dn)
        mat = dn.find(allow_user)
        if mat > -1:
            matched = True
            break

    if not matched:
        return False, "User %s is not allowed" % str(dn), None

    if matched:
        # username = get_user_name_from_dn(dn)
        ban_user_list = X509Authentication().get_ban_user_list()
        for ban_user in ban_user_list:
            pat = re.compile(ban_user)
            mat = pat.match(dn)
            if mat:
                return False, "User %s is banned" % str(dn), None
    username = get_user_name_from_dn(dn)
    return True, None, username


def authenticate_oidc(vo, token):
    oidc_auth = OIDCAuthentication()
    status, data, username = oidc_auth.verify_id_token(vo, token)
    if status:
        return status, data, username
    else:
        return status, data, username


def authenticate_is_super_user(username, dn=None):
    super_user_list = X509Authentication().get_super_user_list()
    for super_user in super_user_list:
        if username == super_user:
            return True
        if dn and super_user in dn:
            return True
    return False



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2021


import datetime
import logging
import json
import os
import re
import requests
import subprocess
import sys
import tarfile
# import traceback

from enum import Enum
from functools import wraps
from packaging import version as packaging_version

from idds.common.config import (config_has_section, config_has_option,
                                config_get, config_get_bool)
from idds.common.constants import (IDDSEnum, RequestType, RequestStatus,
                                   TransformType, TransformStatus,
                                   CollectionType, CollectionRelationType, CollectionStatus,
                                   ContentType, ContentStatus,
                                   GranularityType, ProcessingStatus)
from idds.common.dict_class import DictClass
from idds.common.exceptions import IDDSException


# RFC 1123
DATE_FORMAT = '%a, %d %b %Y %H:%M:%S UTC'


def setup_logging(name, stream=None, loglevel=None):
    """
    Setup logging
    """
    if loglevel is None:
        if config_has_section('common') and config_has_option('common', 'loglevel'):
            loglevel = getattr(logging, config_get('common', 'loglevel').upper())
        else:
            loglevel = logging.INFO

    if stream is None:
        if config_has_section('common') and config_has_option('common', 'logdir'):
            logging.basicConfig(filename=os.path.join(config_get('common', 'logdir'), name),
                                level=loglevel,
                                format='%(asctime)s\t%(threadName)s\t%(name)s\t%(levelname)s\t%(message)s')
        else:
            logging.basicConfig(stream=sys.stdout, level=loglevel,
                                format='%(asctime)s\t%(threadName)s\t%(name)s\t%(levelname)s\t%(message)s')
    else:
        logging.basicConfig(stream=stream, level=loglevel,
                            format='%(asctime)s\t%(threadName)s\t%(name)s\t%(levelname)s\t%(message)s')


def get_rest_url_prefix():
    if config_has_section('rest') and config_has_option('rest', 'url_prefix'):
        url_prefix = config_get('rest', 'url_prefix')
    else:
        url_prefix = None
    if url_prefix:
        while url_prefix.startswith('/'):
            url_prefix = url_prefix[1:]
        while url_prefix.endswith('/'):
            url_prefix = url_prefix[:-1]
        url_prefix = '/' + url_prefix
    return url_prefix


def get_rest_debug():
    if config_has_section('rest') and config_has_option('rest', 'debug'):
        return config_get_bool('rest', 'debug')
    return False


def get_rest_cacher_dir():
    cacher_dir = None
    if config_has_section('rest') and config_has_option('rest', 'cacher_dir'):
        cacher_dir = config_get('rest', 'cacher_dir')
    if cacher_dir and os.path.exists(cacher_dir):
        return cacher_dir
    raise Exception("cacher_dir is not defined or it doesn't exist")


def str_to_date(string):
    """
    Converts a string to the corresponding datetime value.

    :param string: the string to convert to datetime value.
    """
    return datetime.datetime.strptime(string, DATE_FORMAT) if string else None


def date_to_str(date):
    """
    Converts a datetime value to a string.

    :param date: the datetime value to convert.
    """
    return datetime.datetime.strftime(date, DATE_FORMAT) if date else None


def has_config():
    """
    check whether there is a config file
    """
    if os.environ.get('IDDS_CONFIG', None):
        configfile = os.environ.get('IDDS_CONFIG', None)
        if configfile and os.path.exists(configfile):
            return True
    else:
        configfiles = ['%s/etc/idds/idds.cfg' % os.environ.get('IDDS_HOME', ''),
                       '/etc/idds/idds.cfg',
                       '%s/etc/idds/idds.cfg' % os.environ.get('VIRTUAL_ENV', '')]

        for configfile in configfiles:
            if configfile and os.path.exists(configfile):
                return True
    return False


def check_rest_host():
    """
    Function to check whether rest host is defined in config.
    To be used to decide whether to skip some test functions.

    :returns True: if rest host is available. Otherwise False.
    """
    if config_has_option('rest', 'host'):
        host = config_get('rest', 'host')
        if host:
            return True
    return False


def get_rest_host():
    """
    Function to get rest host
    """
    if "IDDS_HOST" in os.environ:
        return os.environ.get("IDDS_HOST")
    host = config_get('rest', 'host')
    url_prefix = get_rest_url_prefix()
    while host.endswith("/"):
        host = host[:-1]
    if url_prefix:
        host = ''.join([host, url_prefix])
    return host


def check_user_proxy():
    """
    Check whether there is a user proxy.
    """
    if 'X509_USER_PROXY' in os.environ:
        client_proxy = os.environ['X509_USER_PROXY']
    else:
        client_proxy = '/tmp/x509up_u%d' % os.geteuid()

    if not os.path.exists(client_proxy):
        return False
    else:
        return True


def check_database():
    """
    Function to check whether database is defined in config.
    To be used to decide whether to skip some test functions.

    :returns True: if database.default is available. Otherwise False.
    """
    if config_has_option('database', 'default'):
        database = config_get('database', 'default')
        if database:
            return True
    return False


def run_process(cmd, stdout=None, stderr=None):
    """
    Runs a command in an out-of-procees shell.
    """
    if stdout and stderr:
        process = subprocess.Popen(cmd, shell=True, stdout=stdout, stderr=stderr, preexec_fn=os.setsid)
    else:
        process = subprocess.Popen(cmd, shell=True)
    return process


def run_command(cmd):
    """
    Runs a command in an out-of-procees shell.
    """
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, preexec_fn=os.setsid)
    stdout, stderr = process.communicate()
    if stdout is not None and type(stdout) in [bytes]:
        stdout = stdout.decode()
    if stderr is not None and type(stderr) in [bytes]:
        stderr = stderr.decode()
    status = process.returncode
    return status, stdout, stderr


def get_space_from_string(space_str):
    """
    Convert space with P, T, G, M to int
    """
    M = 1024
    G = 1024 * M
    T = 1024 * G
    P = 1024 * T

    if 'M' in space_str:
        return int(float(space_str.split('M')[0]) * M)
    elif 'G' in space_str:
        return int(float(space_str.split('G')[0]) * G)
    elif 'T' in space_str:
        return int(float(space_str.split('T')[0]) * T)
    elif 'P' in space_str:
        return int(float(space_str.split('P')[0]) * P)
    else:
        return int(space_str)


def urlretrieve(url, dest, timeout=300):
    """
    Download a file.

    :param url: The url of the source file.
    :param dest: destination file path.
    """
    with open(dest, 'wb') as f:
        r = requests.get(url, allow_redirects=True, timeout=timeout)
        if r.status_code == 200:
            f.write(r.content)
            return 0
        else:
            return -1


def convert_nojsontype_to_value(params):
    """
    Convert enum to its value

    :param params: dict of parameters.

    :returns: dict of parameters.
    """
    if isinstance(params, list):
        new_params = []
        for v in params:
            if v is not None:
                if isinstance(v, Enum):
                    v = v.value
                if isinstance(v, datetime.datetime):
                    v = date_to_str(v)
                if isinstance(v, (list, dict)):
                    v = convert_nojsontype_to_value(v)
            new_params.append(v)
        params = new_params
    elif isinstance(params, dict):
        for key in params:
            if params[key] is not None:
                if isinstance(params[key], Enum):
                    params[key] = params[key].value
                if isinstance(params[key], datetime.datetime):
                    params[key] = date_to_str(params[key])
                if isinstance(params[key], (list, dict)):
                    params[key] = convert_nojsontype_to_value(params[key])
    return params


def convert_value_to_nojsontype(params):
    """
    Convert value to enum

    :param params: dict of parameters.

    :returns: dict of parameters.
    """
    req_keys = {'request_type': RequestType, 'status': RequestStatus}
    transform_keys = {'transform_type': TransformType, 'status': TransformStatus}
    coll_keys = {'coll_type': CollectionType, 'relation_type': CollectionRelationType, 'coll_status': CollectionStatus}
    content_keys = {'content_type': ContentType, 'status': ContentStatus}
    process_keys = {'granularity_type': GranularityType, 'status': ProcessingStatus}

    if 'request_type' in params:
        keys = req_keys
    elif 'transform_type' in params:
        keys = transform_keys
    elif 'coll_type' in params:
        keys = coll_keys
    elif 'content_type' in params:
        keys = content_keys
    elif 'granularity_type' in params:
        keys = process_keys

    if isinstance(params, list):
        new_params = []
        for v in params:
            if v is not None and isinstance(v, (list, dict)):
                v = convert_value_to_nojsontype(v)
            new_params.append(v)
        params = new_params
    elif isinstance(params, dict):
        keys = []
        if 'request_type' in params:
            keys = req_keys
        elif 'transform_type' in params:
            keys = transform_keys
        elif 'coll_type' in params:
            keys = coll_keys
        elif 'content_type' in params:
            keys = content_keys
        elif 'granularity_type' in params:
            keys = process_keys

        for key in keys.keys():
            if key in params and params[key] is not None and isinstance(params[key], int):
                params[key] = keys[key](params[key])

        for key in params:
            if params[key] is not None:
                if isinstance(params[key], (list, dict)):
                    params[key] = convert_value_to_nojsontype(params[key])

    return params


def convert_request_type_to_transform_type(request_type):
    if isinstance(request_type, RequestType):
        request_type = request_type.value
    return TransformType(request_type)


class DictClassEncoder(json.JSONEncoder):
    def default(self, obj):
        # print(obj)
        if isinstance(obj, IDDSEnum) or isinstance(obj, DictClass):
            return obj.to_dict()
        elif isinstance(obj, datetime.datetime):
            return date_to_str(obj)
        elif isinstance(obj, datetime.timedelta):
            return str(obj)
        # elif isinstance(obj, (datetime.time, datetime.date)):
        #     return obj.isoformat()
        # elif isinstance(obj, datetime.timedelta):
        #     return obj.days * 24 * 60 * 60 + obj.seconds

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def as_has_dict(dct):
    if DictClass.is_class(dct):
        return DictClass.from_dict(dct)
    return dct


def json_dumps(obj, indent=None, sort_keys=False):
    return json.dumps(obj, indent=indent, sort_keys=sort_keys, cls=DictClassEncoder)


def json_loads(obj):
    return json.loads(obj, object_hook=as_has_dict)


def get_parameters_from_string(text):
    """
    Find all strings starting with '%'. For example, for this string below, it should return ['NUM_POINTS', 'IN', 'OUT']
    'run --rm -it -v "$(pwd)":/payload gitlab-registry.cern.ch/zhangruihpc/endpointcontainer:latest /bin/bash -c "echo "--num_points %NUM_POINTS"; /bin/cat /payload/%IN>/payload/%OUT"'
    """
    ret = re.findall(r"[%]\w+", text)
    ret = [r.replace('%', '') for r in ret]
    # remove dumplications
    ret = list(set(ret))
    return ret


def replace_parameters_with_values(text, values):
    """
    Replace all strings starting with '%'. For example, for this string below, it should replace ['%NUM_POINTS', '%IN', '%OUT']
    'run --rm -it -v "$(pwd)":/payload gitlab-registry.cern.ch/zhangruihpc/endpointcontainer:latest /bin/bash -c "echo "--num_points %NUM_POINTS"; /bin/cat /payload/%IN>/payload/%OUT"'

    :param text
    :param values: parameter values, for example {'NUM_POINTS': 5, 'IN': 'input.json', 'OUT': 'output.json'}
    """
    for key in values:
        key1 = '%' + key
        text = re.sub(key1, str(values[key]), text)
    return text


def tar_zip_files(output_dir, output_filename, files):
    output_filename = os.path.join(output_dir, output_filename)
    with tarfile.open(output_filename, "w:gz") as tar:
        for file in files:
            tar.add(file, arcname=os.path.basename(file))


def exception_handler(function):
    @wraps(function)
    def new_funct(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except IDDSException as ex:
            logging.error(ex)
            # print(traceback.format_exc())
            return str(ex)
        except Exception as ex:
            logging.error(ex)
            # print(traceback.format_exc())
            return str(ex)
    return new_funct


def is_sub(a, b):
    if not a:
        return True

    for i in a:
        if i not in b:
            return False
    return True


def get_proxy_path():
    try:
        if 'X509_USER_PROXY' in os.environ:
            proxy = os.environ['X509_USER_PROXY']
            if os.path.exists(proxy) and os.access(proxy, os.R_OK):
                return proxy
        proxy = '/tmp/x509up_u%s' % os.getuid()
        if os.path.exists(proxy) and os.access(proxy, os.R_OK):
            return proxy
    except Exception as ex:
        raise IDDSException("Cannot find User proxy: %s" % str(ex))
    return None


def get_proxy():
    try:
        proxy = get_proxy_path()
        if not proxy:
            return proxy
        with open(proxy, 'r') as fp:
            data = fp.read()
        return data
    except Exception as ex:
        raise IDDSException("Cannot find User proxy: %s" % str(ex))
    return None


def is_new_version(version1, version2):
    return packaging_version.parse(version1) > packaging_version.parse(version2)


def extract_scope_atlas(did, scopes):
    # Try to extract the scope from the DSN
    if did.find(':') > -1:
        if len(did.split(':')) > 2:
            raise IDDSException('Too many colons. Cannot extract scope and name')
        scope, name = did.split(':')[0], did.split(':')[1]
        if name.endswith('/'):
            name = name[:-1]
        return scope, name
    else:
        scope = did.split('.')[0]
        if did.startswith('user') or did.startswith('group'):
            scope = ".".join(did.split('.')[0:2])
        if did.endswith('/'):
            did = did[:-1]
        return scope, did


def truncate_string(string, length=800):
    string = (string[:length] + '...') if string and len(string) > length else string
    return string



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
Class of base plugin
"""

import logging

from idds.common import exceptions


class PluginBase(object):
    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

        self.logger = None
        self.setup_logger()
        self.plugins = self.load_plugins(kwargs)

    def get_class_name(self):
        return self.__class__.__name__

    def setup_logger(self, logger=None):
        """
        Setup logger
        """
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(self.get_class_name())

    def set_logger(self, logger):
        self.logger = logger

    def __call__(self, **kwargs):
        return exceptions.NotImplementedException(self.get_class_name())

    def load_plugin_attributes(self, name, plugin, kwargs):
        """
        Load plugin attributes
        """
        attrs = {}
        for option, value in kwargs.items():
            plugin_prefix = 'plugin.%s.' % name
            if option.startswith(plugin_prefix):
                attr_name = option.replace(plugin_prefix, '')
                if isinstance(value, str) and value.lower() == 'true':
                    value = True
                if isinstance(value, str) and value.lower() == 'false':
                    value = False
                attrs[attr_name] = value
        return attrs

    def load_plugin(self, name, plugin, kwargs):
        """
        Load plugin
        """
        attrs = self.load_plugin_attributes(name, plugin, kwargs)
        k = plugin.rfind('.')
        plugin_modules = plugin[:k]
        plugin_class = plugin[k + 1:]
        module = __import__(plugin_modules, fromlist=[None])
        cls = getattr(module, plugin_class)
        impl = cls(**attrs)
        return impl

    def load_plugins(self, kwargs):
        if not kwargs:
            return {}

        plugins = {}
        for key, value in kwargs.items():
            if key.startswith('plugin.'):
                if key.count('.') == 1:
                    plugin_name = key.replace('plugin.', '').strip()
                    plugins[plugin_name] = self.load_plugin(plugin_name, value, kwargs)
        return plugins



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022

import json
from dogpile.cache import make_region

from idds.common.config import config_has_section, config_has_option, config_get


def get_cache_url():
    if config_has_section('cache') and config_has_option('cache', 'url'):
        return config_get('cache', 'url')
    return '127.0.0.1:11211'


def make_region_memcached(expiration_time, function_key_generator=None):
    """
    Make and configure a dogpile.cache.memcached region
    """
    if function_key_generator:
        region = make_region(function_key_generator=function_key_generator)
    else:
        region = make_region()

    region.configure(
        'dogpile.cache.memcached',
        expiration_time=expiration_time,
        arguments={
            'url': get_cache_url,
            'distributed_lock': True,
            'memcached_expire_time': expiration_time + 60,  # must be bigger than expiration_time
        }
    )

    return region


REGION = make_region_memcached(expiration_time=3600)


def update_cache(key, data):
    REGION.set(key, json.dumps(data))


def get_cache(key):
    data = REGION.get(key)
    if data:
        return json.loads(data)
    return data


def delete_cache(key):
    REGION.delete(key)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019

"""
IDDS Exceptions.

error codes:
    The fist the number is one of the main catagories.
    The second number is one of the subcatagories.
    The third number and numbers after third one are local defined for every subcatagory.

Catagories:
 1. common/unknown IDDS exception
 2. ORM related exception
    1. request table related exception
    2. transform table related exception
    3. collection table related exception
    4. collection_content table related exception
    5. processing table related exception
    6. messages table related exception
 3. core related exception
 4. Rest related exception
    1. bad request
    2. connection error
 5. Agent exception
    1. No plugin exception
    2. Agent plugin exception
"""


import traceback


class IDDSException(Exception):
    """
    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """

    def __init__(self, *args, **kwargs):
        super(IDDSException, self).__init__()
        self._message = "An unknown IDDS exception occurred."
        self.args = args
        self.kwargs = kwargs
        self.error_code = 100
        self._error_string = None

    def construct_error_string(self):
        try:
            self._error_string = "%s: %s" % (self._message, self.kwargs)
        except Exception:
            # at least get the core message out if something happened
            self._error_string = self._message
        if len(self.args) > 0:
            # Convert all arguments into their string representations...
            args = ["%s" % arg for arg in self.args if arg]
            self._error_string = (self._error_string + "\nDetails: %s" % '\n'.join(args))
        return self._error_string.strip()

    def __str__(self):
        self.construct_error_string()
        return self._error_string.strip()

    def get_detail(self):
        self.construct_error_string()
        return self._error_string.strip() + "\nStacktrace: %s" % traceback.format_exc()


class NotImplementedException(IDDSException):
    """
    RucioException
    """
    def __init__(self, *args, **kwargs):
        super(NotImplementedException, self).__init__(*args, **kwargs)
        self._message = "Not implemented exception."
        self.error_code = 101


class WrongParameterException(IDDSException):
    """
    RucioException
    """
    def __init__(self, *args, **kwargs):
        super(WrongParameterException, self).__init__(*args, **kwargs)
        self._message = "Wrong parameter exception."
        self.error_code = 102


class DatabaseException(IDDSException):
    """
    RucioException
    """
    def __init__(self, *args, **kwargs):
        super(DatabaseException, self).__init__(*args, **kwargs)
        self._message = "Database exception."
        self.error_code = 200


class InvalidDatabaseType(DatabaseException):
    """
    RucioException
    """
    def __init__(self, *args, **kwargs):
        super(InvalidDatabaseType, self).__init__(*args, **kwargs)
        self._message = "Invalid Database type exception."
        self.error_code = 201


class NoObject(DatabaseException):
    """
    RucioException
    """
    def __init__(self, *args, **kwargs):
        super(NoObject, self).__init__(*args, **kwargs)
        self._message = "No object  exception."
        self.error_code = 202


class DuplicatedObject(DatabaseException):
    """
    RucioException
    """
    def __init__(self, *args, **kwargs):
        super(DuplicatedObject, self).__init__(*args, **kwargs)
        self._message = "Duplicated object exception."
        self.error_code = 203


class CoreException(IDDSException):
    """
    CoreException
    """
    def __init__(self, *args, **kwargs):
        super(CoreException, self).__init__(*args, **kwargs)
        self._message = "Core exception."
        self.error_code = 300


class ConflictRequestException(CoreException):
    """
    ConflictRequestException
    """
    def __init__(self, *args, **kwargs):
        super(ConflictRequestException, self).__init__(*args, **kwargs)
        self._message = "Conflict Reqeust exception."
        self.error_code = 301


class RestException(IDDSException):
    """
    RestException
    """
    def __init__(self, *args, **kwargs):
        super(RestException, self).__init__(*args, **kwargs)
        self._message = "Rest exception."
        self.error_code = 400


class BadRequest(RestException):
    """
    BadRequest
    """
    def __init__(self, *args, **kwargs):
        super(BadRequest, self).__init__(*args, **kwargs)
        self._message = "Bad request exception."
        self.error_code = 401


class ConnectionException(RestException):
    """
    ConnectionException
    """
    def __init__(self, *args, **kwargs):
        super(ConnectionException, self).__init__(*args, **kwargs)
        self._message = "Connection exception."
        self.error_code = 402


class ProcessNotFound(IDDSException):
    """
    ProcessNotFound
    """
    def __init__(self, *args, **kwargs):
        super(ProcessNotFound, self).__init__(*args, **kwargs)
        self._message = "Process not found."
        self.error_code = 403


class ProcessFormatNotSupported(IDDSException):
    """
    ProcessFormatNotSupported
    """
    def __init__(self, *args, **kwargs):
        super(ProcessFormatNotSupported, self).__init__(*args, **kwargs)
        self._message = "Process format not support."
        self.error_code = 404


class ProcessSubmitFailed(IDDSException):
    """
    ProcessSubmitFailed
    """
    def __init__(self, *args, **kwargs):
        super(ProcessSubmitFailed, self).__init__(*args, **kwargs)
        self._message = "Failed to submit process."
        self.error_code = 405


class AgentException(IDDSException):
    """
    BrokerException
    """
    def __init__(self, *args, **kwargs):
        super(AgentException, self).__init__(*args, **kwargs)
        self._message = "Agent exception."
        self.error_code = 500


class NoPluginException(AgentException):
    """
    NoPluginException exception
    """
    def __init__(self, *args, **kwargs):
        super(NoPluginException, self).__init__(*args, **kwargs)
        self._message = "No plugin exception."
        self.error_code = 501


class AgentPluginError(AgentException):
    """
    AgentPluginError exception
    """
    def __init__(self, *args, **kwargs):
        super(AgentPluginError, self).__init__(*args, **kwargs)
        self._message = "Agent plugin exception."
        self.error_code = 502


class AuthenticationException(IDDSException):
    """
    AuthenticationException
    """
    def __init__(self, *args, **kwargs):
        super(AuthenticationException, self).__init__(*args, **kwargs)
        self._message = "Authentication exception."
        self.error_code = 600


class AuthenticationPending(IDDSException):
    """
    Authentication pending
    """
    def __init__(self, *args, **kwargs):
        super(AuthenticationPending, self).__init__(*args, **kwargs)
        self._message = "Authentication pending."
        self.error_code = 601


class AuthenticationNotSupported(IDDSException):
    """
    Authentication not supported
    """
    def __init__(self, *args, **kwargs):
        super(AuthenticationNotSupported, self).__init__(*args, **kwargs)
        self._message = "Authentication not supported."
        self.error_code = 602


class AuthenticationNoPermission(IDDSException):
    """
    Authentication No Permission
    """
    def __init__(self, *args, **kwargs):
        super(AuthenticationNoPermission, self).__init__(*args, **kwargs)
        self._message = "Authentication no permission."
        self.error_code = 603
